"""
Proper Scoring Rules for assessing the quality of predictive
uncertainty quantification.
"""

import numpy as np
from scipy import stats
from uncertainty_toolbox.utils import assert_is_flat_same_shape


def nll_gaussian(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    scaled: bool = True,
) -> float:
    """Negative log likelihood for a gaussian.

    The negative log likelihood for held out data (y_true) given predictive
    uncertainty with mean (y_pred) and standard-deviation (y_std).

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        scaled: Whether to scale the negative log likelihood by size of held out set.

    Returns:
        The negative log likelihood for the heldout set.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)

    # Set residuals
    residuals = y_pred - y_true

    # Compute nll
    nll_list = stats.norm.logpdf(residuals, scale=y_std)
    nll = -1 * np.sum(nll_list)

    # Potentially scale so that sum becomes mean
    if scaled:
        nll = nll / len(nll_list)

    return nll


def crps_gaussian(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    scaled: bool = True,
) -> float:
    """The negatively oriented continuous ranked probability score for Gaussians.

    Computes CRPS for held out data (y_true) given predictive uncertainty with mean
    (y_pred) and standard-deviation (y_std). Each test point is given equal weight
    in the overall score over the test set.

    Negatively oriented means a smaller value is more desirable.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        scaled: Whether to scale the score by size of held out set.

    Returns:
        The crps for the heldout set.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)

    # Compute crps
    y_standardized = (y_true - y_pred) / y_std
    term_1 = 1 / np.sqrt(np.pi)
    term_2 = 2 * stats.norm.pdf(y_standardized, loc=0, scale=1)
    term_3 = y_standardized * (2 * stats.norm.cdf(y_standardized, loc=0, scale=1) - 1)

    crps_list = -1 * y_std * (term_1 - term_2 - term_3)
    crps = np.sum(crps_list)

    # Potentially scale so that sum becomes mean
    if scaled:
        crps = crps / len(crps_list)

    return crps


def check_score(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    scaled: bool = True,
    start_q: float = 0.01,
    end_q: float = 0.99,
    resolution: int = 99,
) -> float:
    """The negatively oriented check score.

    Computes the negatively oriented check score for held out data (y_true)
    given predictive uncertainty with mean (y_pred) and standard-deviation (y_std).
    Each test point and each quantile is given equal weight in the overall score
    over the test set and list of quantiles.

    The score is computed by scanning over a sequence of quantiles of the predicted
    distributions, starting at (start_q) and ending at (end_q).

    Negatively oriented means a smaller value is more desirable.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        scaled: Whether to scale the score by size of held out set.
        start_q: The lower bound of the quantiles to use for computation.
        end_q: The upper bound of the quantiles to use for computation.
        resolution: The number of quantiles to use for computation.

    Returns:
        The check score.
    """
    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)

    test_qs = np.linspace(start_q, end_q, resolution)

    check_list = []
    for q in test_qs:
        q_level = stats.norm.ppf(q, loc=y_pred, scale=y_std)  # pred quantile
        diff = q_level - y_true
        mask = (diff >= 0).astype(float) - q
        score_per_q = np.mean(mask * diff)
        check_list.append(score_per_q)
    check_score = np.sum(check_list)

    if scaled:
        check_score = check_score / len(check_list)

    return check_score


def interval_score(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    scaled: bool = True,
    start_p: float = 0.01,
    end_p: float = 0.99,
    resolution: int = 99,
) -> float:
    """The negatively oriented interval score.

    Compute the negatively oriented interval score for held out data (y_true)
    given predictive uncertainty with mean (y_pred) and standard-deviation
    (y_std). Each test point and each percentile is given equal weight in the
    overall score over the test set and list of quantiles.

    Negatively oriented means a smaller value is more desirable.

    This metric is computed by scanning over a sequence of prediction intervals. Where
    p is the amount of probability captured from a centered prediction interval,
    intervals are formed starting at p=(start_p) and ending at p=(end_p).

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        scaled: Whether to scale the score by size of held out set.
        start_p: The lower bound of probability to capture in a prediction interval.
        end_p: The upper bound of probability to capture in a prediction interval.
        resolution: The number of prediction intervals to use to compute the metric.

    Returns:
        The interval score.
    """
    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)

    test_ps = np.linspace(start_p, end_p, resolution)

    int_list = []
    for p in test_ps:
        low_p, high_p = 0.5 - (p / 2.0), 0.5 + (p / 2.0)  # p% PI
        pred_l = stats.norm.ppf(low_p, loc=y_pred, scale=y_std)
        pred_u = stats.norm.ppf(high_p, loc=y_pred, scale=y_std)

        below_l = ((pred_l - y_true) > 0).astype(float)
        above_u = ((y_true - pred_u) > 0).astype(float)

        score_per_p = (
            (pred_u - pred_l)
            + (2.0 / (1 - p)) * (pred_l - y_true) * below_l
            + (2.0 / (1 - p)) * (y_true - pred_u) * above_u
        )
        mean_score_per_p = np.mean(score_per_p)
        int_list.append(mean_score_per_p)
    int_score = np.sum(int_list)

    if scaled:
        int_score = int_score / len(int_list)

    return int_score



"""
Code for importing and generating data.
"""
from typing import Tuple

import numpy as np


def synthetic_arange_random(
    num_points: int = 10,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Dataset of evenly spaced points and identity function (with some randomization).

    This function returns predictions and predictive uncertainties (given as standard
    deviations) from some hypothetical uncertainty model, along with true input x and
    output y data points.

    Args:
        num_points: The number of data points in the set.

    Returns:
        - The y predictions given by a hypothetical predictive uncertainty model. These
          are the true values of y but with uniform noise added.
        - The standard deviations given by a hypothetical predictive uncertainty model.
          These are the errors between the predictions and the truth plus some unifom
          noise.
        - The true outputs y.
        - The true inputs x.
    """
    x = np.arange(num_points)
    y_true = np.arange(num_points)
    y_pred = np.arange(num_points) + np.random.random((num_points,))
    y_std = np.abs(y_true - y_pred) + 0.1 * np.random.random((num_points,))

    return y_pred, y_std, y_true, x


def synthetic_sine_heteroscedastic(
    n_points: int = 10,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Return samples from "synthetic sine" heteroscedastic noisy function.

    This returns a synthetic dataset which can be used to train and assess a predictive
    uncertainty model.

    Args:
        n_points: The number of data points in the set.

    Returns:
        - Predicted output points y.
        - Predictive uncertainties, defined using standard deviation of added noise.
        - True output points y.
        - True input points x.
    """
    bounds = [0, 15]

    x = np.linspace(bounds[0], bounds[1], n_points)

    f = np.sin(x)
    std = 0.01 + np.abs(x - 5.0) / 10.0
    noise = np.random.normal(scale=std)
    y = f + noise
    return f, std, y, x



"""
Util functions for the toolbox.
"""
from typing import Any, NoReturn, Union

import numpy as np


def assert_is_flat_same_shape(*args: Any) -> Union[bool, NoReturn]:
    """Check if inputs are all same-length 1d numpy.ndarray.

    Args:
        args: the numpy arrays to check.

    Returns:
        True if all arrays are flat and the same shape, or else raises assertion error.
    """

    assert isinstance(args[0], np.ndarray), "All inputs must be of type numpy.ndarray"
    first_shape = args[0].shape
    for arr in args:
        assert isinstance(arr, np.ndarray), "All inputs must be of type numpy.ndarray"
        assert len(arr.shape) == 1, "All inputs must be 1d numpy.ndarray"
        assert arr.shape == first_shape, "All inputs must have the same length"

    return True


def assert_is_positive(*args: Any) -> Union[bool, NoReturn]:
    """Assert that all numpy arrays are positive.

    Args:
        args: the numpy arrays to check.

    Returns:
        True if all elements in all arrays are positive values, or else raises assertion error.
    """
    for arr in args:
        assert isinstance(arr, np.ndarray), "All inputs must be of type numpy.ndarray"
        assert all(arr > 0.0)

    return True



"""
Metrics for assessing the quality of predictive uncertainty quantification.
"""
from typing import Dict

import numpy as np
from sklearn.metrics import (
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    median_absolute_error,
)
from uncertainty_toolbox.utils import assert_is_flat_same_shape


def prediction_error_metrics(
    y_pred: np.ndarray,
    y_true: np.ndarray,
) -> Dict[str, float]:
    """Get all prediction error metrics.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.

    Returns:
        A dictionary with Mean average error ('mae'), Root mean squared
        error ('rmse'), Median absolute error ('mdae'),  Mean absolute
        relative percent difference ('marpd'), r^2 ('r2'), and Pearson's
        correlation coefficient ('corr').
    """
    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_true)

    # Compute metrics
    mae = mean_absolute_error(y_true, y_pred)
    rmse = np.sqrt(mean_squared_error(y_true, y_pred))
    mdae = median_absolute_error(y_true, y_pred)
    residuals = y_true - y_pred
    marpd = np.abs(2 * residuals / (np.abs(y_pred) + np.abs(y_true))).mean() * 100
    r2 = r2_score(y_true, y_pred)
    corr = np.corrcoef(y_true, y_pred)[0, 1]
    prediction_metrics = {
        "mae": mae,
        "rmse": rmse,
        "mdae": mdae,
        "marpd": marpd,
        "r2": r2,
        "corr": corr,
    }

    return prediction_metrics



"""
Test for util functions.
"""

import pytest
import numpy as np

from uncertainty_toolbox.utils import assert_is_flat_same_shape


def test_is_flat_same_shape_wrong_type():
    wrong = [1, 2, 3]
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(wrong)


def test_is_flat_same_shape_wrong_shape():
    wrong = np.arange(9).reshape(3, 3)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(wrong)


def test_is_flat_same_shape_wrong_type_and_shape():
    first = np.arange(3)
    wrong = np.arange(3).reshape(1, 3)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(first, wrong)


def test_is_flat_same_shape_not_all_same():
    first = np.arange(3)
    wrong = np.arange(5)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(first, wrong)


def test_is_flat_same_shape_correct():
    inputs = [np.arange(5) for _ in range(5)]
    assert_is_flat_same_shape(*inputs)



"""
Examples of code for visualizations.
"""

import numpy as np
import matplotlib.pyplot as plt
import uncertainty_toolbox as uct


# Set plot style
uct.viz.set_style()
uct.viz.update_rc("figure.dpi", 150)
uct.viz.update_rc("text.usetex", False)

# Set random seed
np.random.seed(11)

# Generate synthetic predictive uncertainty results
(y_pred, y_std, y_true, x) = uct.synthetic_arange_random(20)

# Plot xy
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_xy(y_pred, y_std, y_true, x, leg_loc=4, ax=ax)
uct.viz.save_figure("viz_minimal_00", "png")
plt.show()

# Plot intervals
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_intervals(y_pred, y_std, y_true, ax=ax)
uct.viz.save_figure("viz_minimal_01", "png")
plt.show()

# Plot intervals_ordered
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_intervals_ordered(y_pred, y_std, y_true, ax=ax)
uct.viz.save_figure("viz_minimal_02", "png")
plt.show()

# Plot calibration
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_calibration(y_pred, y_std, y_true, ax=ax)
uct.viz.save_figure("viz_minimal_03", "png")
plt.show()

# Plot adversarial group calibration
fig, ax = plt.subplots(1, 1, figsize=(6, 4))
uct.plot_adversarial_group_calibration(y_pred, y_std, y_true, ax=ax)
uct.viz.save_figure("viz_minimal_04", "png")
plt.show()

# Plot sharpness
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_sharpness(y_std, ax=ax)
uct.viz.save_figure("viz_minimal_05", "png")
plt.show()

# Plot residuals vs stds
fig, ax = plt.subplots(1, 1, figsize=(5, 5))
uct.plot_residuals_vs_stds(y_pred, y_std, y_true, ax=ax)
uct.viz.save_figure("viz_minimal_06", "png")
plt.show()



"""
Examples of code for recalibration.
"""

import numpy as np
import matplotlib.pyplot as plt
import uncertainty_toolbox as uct


# Set plot style
uct.viz.set_style()
uct.viz.update_rc("text.usetex", False)  # Set to True for system latex
uct.viz.update_rc("font.size", 14)  # Set font size
uct.viz.update_rc("xtick.labelsize", 14)  # Set font size for xaxis tick labels
uct.viz.update_rc("ytick.labelsize", 14)  # Set font size for yaxis tick labels

# Set random seed
np.random.seed(11)

# Generate synthetic predictive uncertainty results
n_obs = 650
f, std, y, x = uct.synthetic_sine_heteroscedastic(n_obs)

# Save figure (set to True to save)
savefig = True

# List of predictive means and standard deviations
pred_mean_list = [f]
pred_std_list = [
    std * 0.5,  # overconfident
    std * 2.0,  # underconfident
]

# ylims for xy plot
ylims_xy = (-2.51, 3.31)

# Loop through, make plots, and compute metrics
for i, pred_mean in enumerate(pred_mean_list):
    for j, pred_std in enumerate(pred_std_list):

        # Before recalibration
        exp_props, obs_props = uct.get_proportion_lists_vectorized(
            pred_mean, pred_std, y
        )
        mace = uct.mean_absolute_calibration_error(
            pred_mean, pred_std, y, recal_model=None
        )
        rmsce = uct.root_mean_squared_calibration_error(
            pred_mean, pred_std, y, recal_model=None
        )
        ma = uct.miscalibration_area(pred_mean, pred_std, y, recal_model=None)
        print("Before Recalibration:  ", end="")
        print("MACE: {:.5f}, RMSCE: {:.5f}, MA: {:.5f}".format(mace, rmsce, ma))

        fig, axes = plt.subplots(1, 2, figsize=(11, 5))
        uct.plot_calibration(
            pred_mean,
            pred_std,
            y,
            exp_props=exp_props,
            obs_props=obs_props,
            ax=axes.flatten()[0],
        )
        uct.plot_xy(
            pred_mean,
            pred_std,
            y,
            x,
            ax=axes.flatten()[1],
            ylims=ylims_xy,
        )

        uct.viz.save_figure(f"before_recal_{j}", "png")

        # After recalibration
        std_recalibrator = uct.recalibration.get_std_recalibrator(
            pred_mean, pred_std, y
        )
        pred_std_recal = std_recalibrator(pred_std)

        mace = uct.mean_absolute_calibration_error(pred_mean, pred_std_recal, y)
        rmsce = uct.root_mean_squared_calibration_error(pred_mean, pred_std_recal, y)
        ma = uct.miscalibration_area(pred_mean, pred_std_recal, y)
        print("After Recalibration:  ", end="")
        print("MACE: {:.5f}, RMSCE: {:.5f}, MA: {:.5f}".format(mace, rmsce, ma))

        fig, axes = plt.subplots(1, 2, figsize=(11, 5))
        uct.plot_calibration(
            pred_mean,
            pred_std_recal,
            y,
            ax=axes.flatten()[0],
        )
        uct.plot_xy(
            pred_mean,
            pred_std_recal,
            y,
            x,
            ax=axes.flatten()[1],
            ylims=ylims_xy,
        )

        uct.viz.save_figure(f"after_recal_{j}", "png")



"""
Tests for calibration metrics.
"""
import numpy as np
import pytest

from uncertainty_toolbox.metrics_calibration import (
    sharpness,
    root_mean_squared_calibration_error,
    mean_absolute_calibration_error,
    adversarial_group_calibration,
    miscalibration_area,
    get_proportion_lists_vectorized,
    get_proportion_lists,
    get_proportion_in_interval,
    get_prediction_interval,
    get_proportion_under_quantile,
    get_quantile,
)


@pytest.fixture
def supply_test_set():
    y_pred = np.array([1, 2, 3])
    y_std = np.array([0.1, 0.5, 1])
    y_true = np.array([1.5, 3, 2])
    return y_pred, y_std, y_true


def test_sharpness_on_test_set(supply_test_set):
    """Test sharpness on the test set for some dummy values."""
    _, test_std, _ = supply_test_set
    assert np.abs(sharpness(test_std) - 0.648074069840786) < 1e-6


def test_root_mean_squared_calibration_error_on_test_set(supply_test_set):
    """Test root mean squared calibration error on some dummy values."""
    test_rmsce_nonvectorized_interval = root_mean_squared_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="interval"
    )
    test_rmsce_vectorized_interval = root_mean_squared_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="interval"
    )
    assert (
        np.abs(test_rmsce_nonvectorized_interval - test_rmsce_vectorized_interval)
        < 1e-6
    )
    assert np.abs(test_rmsce_vectorized_interval - 0.4165757476562379) < 1e-6

    test_rmsce_nonvectorized_quantile = root_mean_squared_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="quantile"
    )
    test_rmsce_vectorized_quantile = root_mean_squared_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="quantile"
    )
    assert (
        np.abs(test_rmsce_nonvectorized_quantile - test_rmsce_vectorized_quantile)
        < 1e-6
    )
    assert np.abs(test_rmsce_vectorized_quantile - 0.30362567774902066) < 1e-6


def test_mean_absolute_calibration_error_on_test_set(supply_test_set):
    """Test mean absolute calibration error on some dummy values."""
    test_mace_nonvectorized_interval = mean_absolute_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="interval"
    )
    test_mace_vectorized_interval = mean_absolute_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="interval"
    )
    assert (
        np.abs(test_mace_nonvectorized_interval - test_mace_vectorized_interval) < 1e-6
    )
    assert np.abs(test_mace_vectorized_interval - 0.3733333333333335) < 1e-6

    test_mace_nonvectorized_quantile = mean_absolute_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="quantile"
    )
    test_mace_vectorized_quantile = mean_absolute_calibration_error(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="quantile"
    )
    assert (
        np.abs(test_mace_nonvectorized_quantile - test_mace_vectorized_quantile) < 1e-6
    )
    assert np.abs(test_mace_vectorized_quantile - 0.23757575757575758) < 1e-6


def test_adversarial_group_calibration_on_test_set(supply_test_set):
    """Test adversarial group calibration on test set for some dummy values."""
    test_out_interval = adversarial_group_calibration(
        *supply_test_set,
        cali_type="mean_abs",
        prop_type="interval",
        num_bins=100,
        num_group_bins=10,
        draw_with_replacement=False,
        num_trials=10,
        num_group_draws=10,
        verbose=False
    )

    assert np.max(np.abs(test_out_interval.group_size - np.linspace(0, 1, 10))) < 1e-6
    assert np.all(test_out_interval.score_mean < 0.5)
    assert np.abs(test_out_interval.score_mean[-1] - 0.3733333333333335) < 1e-6
    assert np.min(test_out_interval.score_stderr) >= 0

    test_out_quantile = adversarial_group_calibration(
        *supply_test_set,
        cali_type="mean_abs",
        prop_type="quantile",
        num_bins=100,
        num_group_bins=10,
        draw_with_replacement=False,
        num_trials=10,
        num_group_draws=10,
        verbose=False
    )

    assert np.max(np.abs(test_out_quantile.group_size - np.linspace(0, 1, 10))) < 1e-6
    assert np.all(test_out_quantile.score_mean < 0.5)
    assert np.abs(test_out_quantile.score_mean[-1] - 0.2375757575757576) < 1e-6
    assert np.min(test_out_quantile.score_stderr) >= 0


def test_miscalibration_area_on_test_set(supply_test_set):
    """Test miscalibration area on some dummy values."""
    test_miscal_area_nonvectorized_interval = miscalibration_area(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="interval"
    )
    test_miscal_area_vectorized_interval = miscalibration_area(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="interval"
    )
    assert (
        np.abs(
            test_miscal_area_nonvectorized_interval
            - test_miscal_area_vectorized_interval
        )
        < 1e-6
    )
    assert np.abs(test_miscal_area_vectorized_interval - 0.37710437710437716) < 1e-6

    test_miscal_area_nonvectorized_quantile = miscalibration_area(
        *supply_test_set,
        num_bins=100,
        vectorized=False,
        recal_model=None,
        prop_type="quantile"
    )
    test_miscal_area_vectorized_quantile = miscalibration_area(
        *supply_test_set,
        num_bins=100,
        vectorized=True,
        recal_model=None,
        prop_type="quantile"
    )
    assert (
        np.abs(
            test_miscal_area_nonvectorized_quantile
            - test_miscal_area_vectorized_quantile
        )
        < 1e-6
    )
    assert np.abs(test_miscal_area_vectorized_quantile - 0.23916245791245788) < 1e-6


def test_vectorization_for_proportion_list_on_test_set(supply_test_set):
    """Test vectorization in get_proportion_lists on the test set for some dummy values."""
    (
        test_exp_props_nonvec_interval,
        test_obs_props_nonvec_interval,
    ) = get_proportion_lists(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="interval"
    )

    (
        test_exp_props_vec_interval,
        test_obs_props_vec_interval,
    ) = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="interval"
    )
    assert (
        np.max(np.abs(test_exp_props_nonvec_interval - test_exp_props_vec_interval))
        < 1e-6
    )
    assert (
        np.max(np.abs(test_obs_props_nonvec_interval - test_obs_props_vec_interval))
        < 1e-6
    )

    (
        test_exp_props_nonvec_quantile,
        test_obs_props_nonvec_quantile,
    ) = get_proportion_lists(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="quantile"
    )

    (
        test_exp_props_vec_quantile,
        test_obs_props_vec_quantile,
    ) = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="quantile"
    )
    assert (
        np.max(np.abs(test_exp_props_nonvec_quantile - test_exp_props_vec_quantile))
        < 1e-6
    )
    assert (
        np.max(np.abs(test_obs_props_nonvec_quantile - test_obs_props_vec_quantile))
        < 1e-6
    )


def test_get_proportion_lists_vectorized_on_test_set(supply_test_set):
    """Test get_proportion_lists_vectorized on the test set for some dummy values."""
    (
        test_exp_props_interval,
        test_obs_props_interval,
    ) = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="interval"
    )
    assert test_exp_props_interval.shape == test_obs_props_interval.shape
    assert (
        np.max(np.abs(np.unique(test_exp_props_interval) - np.linspace(0, 1, 100)))
        < 1e-6
    )
    assert (
        np.max(
            np.abs(
                np.sort(np.unique(test_obs_props_interval))
                - np.array([0.0, 0.33333333, 0.66666667, 1.0])
            )
        )
        < 1e-6
    )

    (
        test_exp_props_quantile,
        test_obs_props_quantile,
    ) = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="quantile"
    )
    assert test_exp_props_quantile.shape == test_obs_props_quantile.shape
    assert (
        np.max(np.abs(np.unique(test_exp_props_quantile) - np.linspace(0, 1, 100)))
        < 1e-6
    )
    assert (
        np.max(
            np.abs(
                np.sort(np.unique(test_obs_props_quantile))
                - np.array([0.0, 0.33333333, 0.66666667, 1.0])
            )
        )
        < 1e-6
    )


def test_get_proportion_lists_on_test_set(supply_test_set):
    """Test get_proportion_lists on the test set for some dummy values."""
    test_exp_props_interval, test_obs_props_interval = get_proportion_lists(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="interval"
    )
    assert len(test_exp_props_interval) == len(test_obs_props_interval)
    assert (
        np.max(np.abs(np.unique(test_exp_props_interval) - np.linspace(0, 1, 100)))
        < 1e-6
    )
    assert (
        np.max(
            np.abs(
                np.sort(np.unique(test_obs_props_interval))
                - np.array([0.0, 0.33333333, 0.66666667, 1.0])
            )
        )
        < 1e-6
    )

    test_exp_props_quantile, test_obs_props_quantile = get_proportion_lists(
        *supply_test_set, num_bins=100, recal_model=None, prop_type="quantile"
    )
    assert len(test_exp_props_quantile) == len(test_obs_props_quantile)
    assert (
        np.max(np.abs(np.unique(test_exp_props_quantile) - np.linspace(0, 1, 100)))
        < 1e-6
    )
    assert (
        np.max(
            np.abs(
                np.sort(np.unique(test_obs_props_quantile))
                - np.array([0.0, 0.33333333, 0.66666667, 1.0])
            )
        )
        < 1e-6
    )


def test_get_proportion_in_interval_on_test_set(supply_test_set):
    """Test get_proportion_in_interval on the test set for some dummy values."""
    test_quantile_value_list = [
        (0.0, 0.0),
        (0.25, 0.0),
        (0.5, 0.0),
        (0.75, 0.3333333333333333),
        (1.0, 1.0),
    ]
    for (test_q, test_val) in test_quantile_value_list:
        assert (
            np.abs(
                get_proportion_in_interval(*supply_test_set, quantile=test_q) - test_val
            )
            < 1e-6
        )


def test_get_proportion_under_quantile_on_test_set(supply_test_set):
    """Test get_proportion_in_interval on the test set for some dummy values."""
    test_quantile_value_list = [
        (0.0, 0.0),
        (0.25, 0.6666666666666666),
        (0.5, 0.6666666666666666),
        (0.75, 0.6666666666666666),
        (1.0, 1.0),
    ]
    for (test_q, test_val) in test_quantile_value_list:
        assert (
            np.abs(
                get_proportion_under_quantile(*supply_test_set, quantile=test_q)
                - test_val
            )
            < 1e-6
        )


def test_get_prediction_interval_on_test_set(supply_test_set):
    """Test get_prediction_interval on the test set for some dummy values."""
    test_quantile_value_list = [
        (
            0.01,
            np.array([1.00125335, 2.00626673, 3.01253347]),
            np.array([0.99874665, 1.99373327, 2.98746653]),
        ),
        (
            0.25,
            np.array([1.03186394, 2.15931968, 3.31863936]),
            np.array([0.96813606, 1.84068032, 2.68136064]),
        ),
        (
            0.50,
            np.array([1.06744898, 2.33724488, 3.67448975]),
            np.array([0.93255102, 1.66275512, 2.32551025]),
        ),
        (
            0.75,
            np.array([1.11503494, 2.57517469, 4.15034938]),
            np.array([0.88496506, 1.42482531, 1.84965062]),
        ),
        (
            0.99,
            np.array([1.25758293, 3.28791465, 5.5758293]),
            np.array([0.74241707, 0.71208535, 0.4241707]),
        ),
    ]

    y_pred, y_std, y_true = supply_test_set

    with pytest.raises(Exception):
        bounds = get_prediction_interval(y_pred, y_std, quantile=0.0, recal_model=None)

    with pytest.raises(Exception):
        bounds = get_prediction_interval(y_pred, y_std, quantile=1.0, recal_model=None)

    for (test_q, test_upper, test_lower) in test_quantile_value_list:
        bounds = get_prediction_interval(
            y_pred, y_std, quantile=test_q, recal_model=None
        )
        upper_bound = bounds.upper
        lower_bound = bounds.lower
        assert np.max(np.abs(upper_bound - test_upper)) < 1e-6
        assert np.max(np.abs(upper_bound - test_upper)) < 1e-6


def test_get_quantile_on_test_set(supply_test_set):
    """Test get_prediction_interval on the test set for some dummy values."""
    test_quantile_value_list = [
        (0.01, np.array([0.76736521, 0.83682606, 0.67365213])),
        (
            0.25,
            np.array([0.93255102, 1.66275512, 2.32551025]),
        ),
        (
            0.50,
            np.array([1.0, 2.0, 3.0]),
        ),
        (
            0.75,
            np.array([1.06744898, 2.33724488, 3.67448975]),
        ),
        (
            0.99,
            np.array([1.23263479, 3.16317394, 5.32634787]),
        ),
    ]

    y_pred, y_std, y_true = supply_test_set

    with pytest.raises(Exception):
        bound = get_quantile(y_pred, y_std, quantile=0.0, recal_model=None)

    with pytest.raises(Exception):
        bound = get_quantile(y_pred, y_std, quantile=1.0, recal_model=None)

    for (test_q, test_bound) in test_quantile_value_list:
        bound = get_quantile(y_pred, y_std, quantile=test_q, recal_model=None)

        assert np.max(np.abs(bound - test_bound)) < 1e-6



"""
Recalibrating uncertainty estimates.
"""

from typing import Callable, Union
import numpy as np
from sklearn.isotonic import IsotonicRegression
from scipy.optimize import minimize_scalar
import uncertainty_toolbox as uct


def get_q_idx(exp_props: np.ndarray, q: float) -> int:
    """Utility function which outputs the array index of an element.

    Gets the (approximate) index of a specified probability value, q, in the expected proportions array.
    Used as a utility function in isotonic regression recalibration.

    Args:
        exp_props: 1D array of expected probabilities.
        q: a specified probability float.

    Returns:
        An index which specifies the (approximate) index of q in exp_props
    """
    num_pts = exp_props.shape[0]
    target_idx = None
    for idx, x in enumerate(exp_props):
        if idx + 1 == num_pts:
            if round(q, 2) == round(float(exp_props[-1]), 2):
                target_idx = exp_props.shape[0] - 1
            break
        if x <= q < exp_props[idx + 1]:
            target_idx = idx
            break
    if target_idx is None:
        raise ValueError("q must be within exp_props")
    return target_idx


def iso_recal(
    exp_props: np.ndarray,
    obs_props: np.ndarray,
) -> IsotonicRegression:
    """Recalibration algorithm based on isotonic regression.

    Fits and outputs an isotonic recalibration model that maps observed
    probabilities to expected probabilities. This mapping proivdes
    the necessary adjustments to produce better calibrated outputs.

    Args:
        exp_props: 1D array of expected probabilities (values must span [0, 1]).
        obs_props: 1D array of observed probabilities.

    Returns:
        An sklearn IsotonicRegression recalibration model.
    """
    # Flatten
    exp_props = exp_props.flatten()
    obs_props = obs_props.flatten()
    min_obs = np.min(obs_props)
    max_obs = np.max(obs_props)

    iso_model = IsotonicRegression(increasing=True, out_of_bounds="clip")
    # just need observed prop values between 0 and 1
    # problematic if min_obs_p > 0 and max_obs_p < 1
    if not (min_obs == 0.0) and (max_obs == 1.0):
        print("Obs props not ideal: from {} to {}".format(min_obs, max_obs))

    exp_0_idx = get_q_idx(exp_props, 0.0)
    exp_1_idx = get_q_idx(exp_props, 1.0)
    within_01 = obs_props[exp_0_idx : exp_1_idx + 1]

    beg_idx, end_idx = None, None
    # Handle beg_idx
    if exp_0_idx != 0:
        min_obs_below = np.min(obs_props[:exp_0_idx])
        min_obs_within = np.min(within_01)
        if min_obs_below < min_obs_within:
            i = exp_0_idx - 1
            while obs_props[i] > min_obs_below:
                i -= 1
            beg_idx = i
        elif np.sum((within_01 == min_obs_within).astype(float)) > 1:
            # multiple minima in within_01 ==> get last min idx
            i = exp_1_idx - 1
            while obs_props[i] > min_obs_within:
                i -= 1
            beg_idx = i
        elif np.sum((within_01 == min_obs_within).astype(float)) == 1:
            beg_idx = int(np.argmin(within_01) + exp_0_idx)
        else:
            raise RuntimeError("Inspect input arrays. Cannot set beginning index.")
    else:
        beg_idx = exp_0_idx

    # Handle end_idx
    if exp_1_idx < obs_props.shape[0] - 1:
        max_obs_above = np.max(obs_props[exp_1_idx + 1 :])
        max_obs_within = np.max(within_01)
        if max_obs_above > max_obs_within:
            i = exp_1_idx + 1
            while obs_props[i] < max_obs_above:
                i += 1
            end_idx = i + 1
        elif np.sum((within_01 == max_obs_within).astype(float)) > 1:
            # multiple minima in within_01 ==> get last min idx
            i = beg_idx
            while obs_props[i] < max_obs_within:
                i += 1
            end_idx = i + 1
        elif np.sum((within_01 == max_obs_within).astype(float)) == 1:
            end_idx = int(exp_0_idx + np.argmax(within_01) + 1)
        else:
            raise RuntimeError("Inspect input arrays. Cannot set ending index.")
    else:
        end_idx = exp_1_idx + 1

    if end_idx <= beg_idx:
        raise RuntimeError("Ending index before beginning index")

    filtered_obs_props = obs_props[beg_idx:end_idx]
    filtered_exp_props = exp_props[beg_idx:end_idx]

    try:
        iso_model = iso_model.fit(filtered_obs_props, filtered_exp_props)
    except Exception:
        raise RuntimeError("Failed to fit isotonic regression model")

    return iso_model


def optimize_recalibration_ratio(
    y_mean: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    criterion: str = "ma_cal",
) -> float:
    """Scale factor which uniformly recalibrates predicted standard deviations.

    Searches via black-box optimization the standard deviation scale factor (opt_ratio)
    which produces the best recalibration, i.e. updated standard deviation
    can be written as opt_ratio * y_std.

    Args:
        y_mean: 1D array of the predicted means for the recalibration dataset.
        y_std: 1D array of the predicted standard deviations for the recalibration dataset.
        y_true: 1D array of the true means for the recalibration dataset.
        criterion: calibration metric to optimize for during recalibration; must be one of {"ma_cal", "rms_cal", "miscal"}.

    Returns:
        A single scalar which optimally recalibrates the predicted standard deviations.
    """
    if criterion == "ma_cal":
        cal_fn = uct.metrics.mean_absolute_calibration_error
        worst_cal = 0.5
    elif criterion == "rms_cal":
        cal_fn = uct.metrics.root_mean_squared_calibration_error
        worst_cal = np.sqrt(1.0 / 3.0)
    elif criterion == "miscal":
        cal_fn = uct.metrics.miscalibration_area
        worst_cal = 0.5
    else:
        raise RuntimeError("Wrong criterion option")

    def obj(ratio):

        # If ratio is 0, return worst-possible calibration metric
        if ratio == 0:
            return worst_cal

        curr_cal = cal_fn(y_mean, ratio * y_std, y_true)
        return curr_cal

    bounds = (1e-3, 1e3)
    result = minimize_scalar(fun=obj, bounds=bounds)
    opt_ratio = result.x

    if not result.success:
        raise Warning("Optimization did not succeed")
        original_cal = cal_fn(y_mean, y_std, y_true)
        ratio_cal = cal_fn(y_mean, opt_ratio * y_std, y_true)
        if ratio_cal > original_cal:
            raise Warning(
                "No better calibration found, no recalibration performed and returning original uncertainties"
            )
            opt_ratio = 1.0

    return opt_ratio


def get_std_recalibrator(
    y_mean: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    criterion: str = "ma_cal",
) -> Callable[[np.ndarray], np.ndarray]:
    """Standard deviation recalibrator.

    Computes the standard deviation recalibration ratio and returns a function
    which takes in an array of uncalibrated standard deviations and returns
    an array of recalibrated standard deviations.

    Args:
        y_mean: 1D array of the predicted means for the recalibration dataset.
        y_std: 1D array of the predicted standard deviations for the recalibration dataset.
        y_true: 1D array of the true means for the recalibration dataset.
        criterion: calibration metric to optimize for during recalibration; must be one of {"ma_cal", "rms_cal", "miscal"}.

    Returns:
        A function which takes uncalibrated standard deviations as input and
        outputs the recalibrated standard deviations.
    """
    std_recal_ratio = optimize_recalibration_ratio(y_mean, y_std, y_true, criterion)

    def std_recalibrator(std_arr):
        return std_recal_ratio * std_arr

    return std_recalibrator


def get_quantile_recalibrator(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
) -> Callable[[np.ndarray, np.ndarray, Union[float, np.ndarray]], np.ndarray]:
    """Quantile recalibrator.

    Fits an isotonic regression recalibration model and returns a function
    which takes in the mean and standard deviation predictions and a specified
    quantile level, and returns the recalibrated quantile.

    Args:
        y_pred: 1D array of the predicted means for the recalibration dataset.
        y_std: 1D array of the predicted standard deviations for the recalibration dataset.
        y_true: 1D array of the true means for the recalibration dataset.

    Returns:
        A function which outputs the recalibrated quantile prediction.
    """
    exp_props, obs_props = uct.get_proportion_lists_vectorized(
        y_pred, y_std, y_true, prop_type="quantile"
    )
    iso_model = iso_recal(exp_props, obs_props)

    def quantile_recalibrator(y_pred, y_std, quantile):
        recal_quantile = uct.metrics_calibration.get_quantile(
            y_pred, y_std, quantile, recal_model=iso_model
        )
        return recal_quantile

    return quantile_recalibrator


def get_interval_recalibrator(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
) -> Callable[[np.ndarray, np.ndarray, Union[float, np.ndarray]], np.ndarray]:
    """Prediction interval recalibrator.

    Fits an isotonic regression recalibration model and returns a function
    which takes in the mean and standard deviation predictions and a specified
    centered interval coverage level, and returns the recalibrated interval.

    Args:
        y_pred: 1D array of the predicted means for the recalibration dataset.
        y_std: 1D array of the predicted standard deviations for the recalibration dataset.
        y_true: 1D array of the true means for the recalibration dataset.

    Returns:
        A function which outputs the recalibrated prediction interval.
    """
    exp_props, obs_props = uct.get_proportion_lists_vectorized(
        y_pred, y_std, y_true, prop_type="interval"
    )
    iso_model = iso_recal(exp_props, obs_props)

    def interval_recalibrator(y_pred, y_std, quantile):
        recal_bounds = uct.metrics_calibration.get_prediction_interval(
            y_pred, y_std, quantile, recal_model=iso_model
        )
        return recal_bounds

    return interval_recalibrator



"""
Tests for scoring rule metrics.
"""
import numpy as np
import pytest

from uncertainty_toolbox.metrics_scoring_rule import (
    nll_gaussian,
    crps_gaussian,
    check_score,
    interval_score,
)


@pytest.fixture
def supply_test_set():
    y_pred = np.array([1, 2, 3])
    y_std = np.array([0.1, 0.5, 1])
    y_true = np.array([1.5, 3, 2])
    return y_pred, y_std, y_true


def test_nll_gaussian_on_test_set(supply_test_set):
    """Test Gaussian NLL on the test set for some dummy values."""
    assert np.abs(nll_gaussian(*supply_test_set) - 4.920361108686675) < 1e-6


def test_nll_gaussian_on_one_pt():
    """Sanity check by testing one point at mean of gaussian."""
    y_pred = np.array([0])
    y_true = np.array([0])
    y_std = np.array([1 / np.sqrt(2 * np.pi)])
    assert np.abs(nll_gaussian(y_pred, y_std, y_true)) < 1e-6


def test_crps_gaussian_on_test_set(supply_test_set):
    """Test CRPS on the test set for some dummy values."""
    assert np.abs(crps_gaussian(*supply_test_set) - 0.59080610693) < 1e-6


def test_check_score_on_test_set(supply_test_set):
    """Test check score on the test set for some dummy values."""
    assert np.abs(check_score(*supply_test_set) - 0.29801437323836477) < 1e-6


def test_check_score_on_one_pt():
    """Sanity check to show that check score is minimized (i.e. 0) if data
    occurs at the exact requested quantile."""
    y_pred = np.array([0])
    y_true = np.array([1])
    y_std = np.array([1])
    score = check_score(
        y_pred=y_pred,
        y_std=y_std,
        y_true=y_true,
        start_q=0.5 + 0.341,
        end_q=0.5 + 0.341,
        resolution=1,
    )
    assert np.abs(score) < 1e-2


def test_interval_score_on_test_set(supply_test_set):
    """Test interval score on the test set for some dummy values."""
    assert np.abs(interval_score(*supply_test_set) - 3.20755700861995) < 1e-6


def test_interval_score_on_one_pt():
    """Sanity check on interval score. For one point in the center of the
    distribution and intervals one standard deviation and two standard
    deviations away, should return ((1 std) * 2 + (2 std) * 2) / 2 = 3.
    """
    y_pred = np.array([0])
    y_true = np.array([0])
    y_std = np.array([1])
    score = interval_score(
        y_pred=y_pred,
        y_std=y_std,
        y_true=y_true,
        start_p=0.682,
        end_p=0.954,
        resolution=2,
    )
    assert np.abs(score - 3) < 1e-2



"""
Tests for metrics.
"""

import pytest
import numpy as np

from uncertainty_toolbox.metrics import (
    get_all_accuracy_metrics,
    get_all_average_calibration,
    get_all_adversarial_group_calibration,
    get_all_sharpness_metrics,
    get_all_scoring_rule_metrics,
    get_all_metrics,
)


@pytest.fixture
def get_test_set():
    y_pred = np.array([1, 2, 3])
    y_std = np.array([0.1, 0.5, 1])
    y_true = np.array([1.5, 3, 2])
    return y_pred, y_std, y_true


def test_get_all_accuracy_metrics_returns(get_test_set):
    """Test if correct accuracy metrics are returned."""
    y_pred, y_std, y_true = get_test_set
    met_dict = get_all_accuracy_metrics(y_pred, y_true)
    met_keys = met_dict.keys()
    assert len(met_keys) == 6

    met_str_list = ["mae", "rmse", "mdae", "marpd", "r2", "corr"]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)


def test_get_all_average_calibration_returns(get_test_set):
    """Test if correct average calibration metrics are returned."""
    n_bins = 20
    met_dict = get_all_average_calibration(*get_test_set, n_bins)
    met_keys = met_dict.keys()
    assert len(met_keys) == 3

    met_str_list = ["rms_cal", "ma_cal", "miscal_area"]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)


def test_get_all_adversarial_group_calibration_returns(get_test_set):
    """Test if correct adversarial group calibration metrics are returned."""
    n_bins = 20
    met_dict = get_all_adversarial_group_calibration(*get_test_set, n_bins)
    met_keys = met_dict.keys()
    assert len(met_keys) == 2

    met_str_list = ["ma_adv_group_cal", "rms_adv_group_cal"]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)

    for met_str in met_str_list:
        inner_dict = met_dict[met_str]
        inner_keys = inner_dict.keys()
        assert len(inner_keys) == 3
        inner_str_list = [
            "group_sizes",
            "adv_group_cali_mean",
            "adv_group_cali_stderr",
        ]
        bool_list = [s in inner_keys for s in inner_str_list]
        assert all(bool_list)


def test_get_all_sharpness_metrics_returns(get_test_set):
    """Test if correct sharpness metrics are returned."""
    y_pred, y_std, y_true = get_test_set
    met_dict = get_all_sharpness_metrics(y_std)
    met_keys = met_dict.keys()
    assert len(met_keys) == 1
    assert "sharp" in met_keys


def test_get_all_scoring_rule_metrics_returns(get_test_set):
    """Test if correct scoring rule metrics are returned."""
    resolution = 99
    scaled = True
    met_dict = get_all_scoring_rule_metrics(*get_test_set, resolution, scaled)
    met_keys = met_dict.keys()
    assert len(met_keys) == 4

    met_str_list = ["nll", "crps", "check", "interval"]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)


def test_get_all_metrics_returns(get_test_set):
    """Test if correct metrics are returned by get_all_metrics function."""
    met_dict = get_all_metrics(*get_test_set)
    met_keys = met_dict.keys()
    assert len(met_keys) == 5

    met_str_list = [
        "accuracy",
        "avg_calibration",
        "adv_group_calibration",
        "sharpness",
        "scoring_rule",
    ]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)



"""
Tests for recalibration procedures.
"""
import numpy as np
import pytest
import random

from uncertainty_toolbox.recalibration import (
    iso_recal,
    optimize_recalibration_ratio,
    get_std_recalibrator,
    get_quantile_recalibrator,
    get_interval_recalibrator,
)
from uncertainty_toolbox.metrics_calibration import (
    root_mean_squared_calibration_error,
    mean_absolute_calibration_error,
    miscalibration_area,
    get_proportion_lists_vectorized,
    get_prediction_interval,
    get_quantile,
)


@pytest.fixture
def supply_test_set():
    y_pred = np.arange(100) / 100.0
    y_std = np.arange(1, 101) / 20.0
    y_true = np.arange(100) / 100.0 + 0.5
    return y_pred, y_std, y_true


def test_recal_model_mace_criterion_on_test_set(supply_test_set):
    """
    Test recalibration on mean absolute calibration error on the test set
    for some dummy values.
    """
    test_mace = mean_absolute_calibration_error(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=None
    )
    test_exp_props, test_obs_props = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None
    )
    recal_model = iso_recal(test_exp_props, test_obs_props)
    recal_test_mace = mean_absolute_calibration_error(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=recal_model
    )
    recal_exp_props = recal_model.predict(test_obs_props)
    assert np.abs(test_mace - 0.24206060606060598) < 1e-6
    assert np.abs(recal_test_mace - 0.003035353535353514) < 1e-6
    for idx in range(1, recal_exp_props.shape[0]):
        assert recal_exp_props[idx - 1] <= recal_exp_props[idx]


def test_recal_model_rmce_criterion_on_test_set(supply_test_set):
    """
    Test recalibration on root mean squared calibration error on the test set
    for some dummy values.
    """
    test_rmsce = root_mean_squared_calibration_error(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=None
    )
    test_exp_props, test_obs_props = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None
    )
    recal_model = iso_recal(test_exp_props, test_obs_props)
    recal_test_rmsce = root_mean_squared_calibration_error(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=recal_model
    )
    recal_exp_props = recal_model.predict(test_obs_props)

    assert np.abs(test_rmsce - 0.28741418862839013) < 1e-6
    assert np.abs(recal_test_rmsce - 0.003981861230030349) < 1e-6
    for idx in range(1, recal_exp_props.shape[0]):
        assert recal_exp_props[idx - 1] <= recal_exp_props[idx]


def test_recal_model_miscal_area_criterion_on_test_set(supply_test_set):
    """
    Test recalibration on miscalibration area on the test set
    for some dummy values.
    """
    test_miscal_area = miscalibration_area(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=None
    )
    test_exp_props, test_obs_props = get_proportion_lists_vectorized(
        *supply_test_set, num_bins=100, recal_model=None
    )
    recal_model = iso_recal(test_exp_props, test_obs_props)
    recal_test_miscal_area = miscalibration_area(
        *supply_test_set, num_bins=100, vectorized=True, recal_model=recal_model
    )
    recal_exp_props = recal_model.predict(test_obs_props)

    assert np.abs(test_miscal_area - 0.24426139657444004) < 1e-6
    assert np.abs(recal_test_miscal_area - 0.0029569160997732057) < 1e-6
    for idx in range(1, recal_exp_props.shape[0]):
        assert recal_exp_props[idx - 1] <= recal_exp_props[idx]


def test_optimize_recalibration_ratio_mace_criterion(supply_test_set):
    """
    Test standard deviation recalibration on mean absolute calibration error
    on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set
    ma_cal_ratio = optimize_recalibration_ratio(
        y_pred, y_std, y_true, criterion="ma_cal"
    )
    recal_ma_cal = mean_absolute_calibration_error(y_pred, ma_cal_ratio * y_std, y_true)
    recal_rms_cal = root_mean_squared_calibration_error(
        y_pred, ma_cal_ratio * y_std, y_true
    )
    recal_miscal = miscalibration_area(y_pred, ma_cal_ratio * y_std, y_true)

    assert np.abs(ma_cal_ratio - 0.33215708813773176) < 1e-6
    assert np.abs(recal_ma_cal - 0.06821616161616162) < 1e-6
    assert np.abs(recal_rms_cal - 0.08800130087804929) < 1e-6
    assert np.abs(recal_miscal - 0.06886262626262629) < 1e-6


def test_optimize_recalibration_ratio_rmce_criterion(supply_test_set):
    """
    Test standard deviation recalibration on root mean squared calibration error
    on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set
    rms_cal_ratio = optimize_recalibration_ratio(
        y_pred, y_std, y_true, criterion="rms_cal"
    )
    recal_ma_cal = mean_absolute_calibration_error(
        y_pred, rms_cal_ratio * y_std, y_true
    )
    recal_rms_cal = root_mean_squared_calibration_error(
        y_pred, rms_cal_ratio * y_std, y_true
    )
    recal_miscal = miscalibration_area(y_pred, rms_cal_ratio * y_std, y_true)

    assert np.abs(rms_cal_ratio - 0.34900989073212507) < 1e-6
    assert np.abs(recal_ma_cal - 0.06945555555555555) < 1e-6
    assert np.abs(recal_rms_cal - 0.08570902541177935) < 1e-6
    assert np.abs(recal_miscal - 0.07011706864564003) < 1e-6


def test_optimize_recalibration_ratio_miscal_area_criterion(supply_test_set):
    """
    Test standard deviation recalibration on miscalibration area
    on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set
    miscal_ratio = optimize_recalibration_ratio(
        y_pred, y_std, y_true, criterion="miscal"
    )
    recal_ma_cal = mean_absolute_calibration_error(y_pred, miscal_ratio * y_std, y_true)
    recal_rms_cal = root_mean_squared_calibration_error(
        y_pred, miscal_ratio * y_std, y_true
    )
    recal_miscal = miscalibration_area(y_pred, miscal_ratio * y_std, y_true)

    assert np.abs(miscal_ratio - 0.3321912522557988) < 1e-6
    assert np.abs(recal_ma_cal - 0.06821616161616162) < 1e-6
    assert np.abs(recal_rms_cal - 0.08800130087804929) < 1e-6
    assert np.abs(recal_miscal - 0.06886262626262629) < 1e-6


def test_get_prediction_interval_recalibrated(supply_test_set):
    """
    Test standard deviation recalibration on miscalibration area
    on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set
    test_exp_props, test_obs_props = get_proportion_lists_vectorized(
        y_pred, y_std, y_true, num_bins=100, recal_model=None
    )
    recal_model = iso_recal(test_exp_props, test_obs_props)

    test_quantile_prop_list = [
        (0.01, 0.0, 0.0),
        (0.25, 0.69, 0.25),
        (0.5, 0.86, 0.5),
        (0.75, 0.92, 0.75),
        (0.99, 0.97, 0.97),
    ]

    for (q, test_orig_prop, test_recal_prop) in test_quantile_prop_list:
        orig_bounds = get_prediction_interval(y_pred, y_std, q, None)
        recal_bounds = get_prediction_interval(y_pred, y_std, q, recal_model)

        orig_prop = np.mean(
            (orig_bounds.lower <= y_true) * (y_true <= orig_bounds.upper)
        )
        recal_prop = np.mean(
            (recal_bounds.lower <= y_true) * (y_true <= recal_bounds.upper)
        )

        assert np.max(np.abs(test_orig_prop - orig_prop)) < 1e-6
        assert np.max(np.abs(test_recal_prop - recal_prop)) < 1e-6


def test_get_std_recalibrator(supply_test_set):
    """
    Test get_std_recalibration on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set

    test_quantile_prop_list = [
        (0.01, 0.00, 0.00),
        (0.25, 0.06, 0.00),
        (0.50, 0.56, 0.00),
        (0.75, 0.74, 0.56),
        (0.99, 0.89, 0.88),
    ]

    std_recalibrator = get_std_recalibrator(y_pred, y_std, y_true)

    for (q, test_prop_in_pi, test_prop_under_q) in test_quantile_prop_list:
        y_std_recal = std_recalibrator(y_std)
        pi = get_prediction_interval(y_pred, y_std_recal, q)
        prop_in_pi = ((pi.lower <= y_true) * (y_true <= pi.upper)).mean()
        quantile_bound = get_quantile(y_pred, y_std_recal, q)
        prop_under_q = (quantile_bound >= y_true).mean()
        assert np.max(np.abs(test_prop_in_pi - prop_in_pi)) < 1e-6
        assert np.max(np.abs(test_prop_under_q - prop_under_q)) < 1e-6


def test_get_quantile_recalibrator(supply_test_set):
    """
    Test get_std_recalibration on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set

    test_quantile_prop_list = [
        (0.01, 0.00),
        (0.25, 0.00),
        (0.50, 0.00),
        (0.75, 0.00),
        (0.99, 0.83),
    ]

    quantile_recalibrator = get_quantile_recalibrator(y_pred, y_std, y_true)

    for (q, test_prop_under_q) in test_quantile_prop_list:
        quantile_bound_recal = quantile_recalibrator(y_pred, y_std, q)
        assert all(np.isfinite(quantile_bound_recal))
        prop_under_q_recal = (quantile_bound_recal >= y_true).mean()
        assert np.max(np.abs(test_prop_under_q - prop_under_q_recal)) < 1e-6


def test_get_interval_recalibrator(supply_test_set):
    """
    Test get_std_recalibration on the test set for some dummy values.
    """
    random.seed(0)
    np.random.seed(seed=0)

    y_pred, y_std, y_true = supply_test_set

    test_quantile_prop_list = [
        (0.01, 0.00),
        (0.25, 0.25),
        (0.50, 0.50),
        (0.75, 0.75),
        (0.99, 0.97),
    ]

    interval_recalibrator = get_interval_recalibrator(y_pred, y_std, y_true)

    for (q, test_prop_in_interval) in test_quantile_prop_list:
        interval_recal = interval_recalibrator(y_pred, y_std, q)
        prop_in_interval_recal = (
            (interval_recal.lower <= y_true) * (y_true <= interval_recal.upper)
        ).mean()
        assert np.max(np.abs(test_prop_in_interval - prop_in_interval_recal)) < 1e-6



"""
Metrics for assessing the quality of predictive uncertainty quantification.
"""

from typing import Any, Tuple, Optional
from argparse import Namespace
import numpy as np
from sklearn.isotonic import IsotonicRegression
from scipy import stats
from shapely.geometry import Polygon, LineString
from shapely.ops import polygonize, unary_union
from sklearn.isotonic import IsotonicRegression
from tqdm import tqdm
from uncertainty_toolbox.utils import (
    assert_is_flat_same_shape,
    assert_is_positive,
)


def sharpness(y_std: np.ndarray) -> float:
    """Return sharpness (a single measure of the overall confidence).

    Args:
        y_std: 1D array of the predicted standard deviations for the held out dataset.

    Returns:
        A single scalar which quantifies the average of the standard deviations.
    """
    # Check that input arrays are flat
    assert_is_flat_same_shape(y_std)
    # Check that input std is positive
    assert_is_positive(y_std)

    # Compute sharpness
    sharp_metric = np.sqrt(np.mean(y_std ** 2))

    return sharp_metric


def root_mean_squared_calibration_error(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    vectorized: bool = False,
    recal_model: IsotonicRegression = None,
    prop_type: str = "interval",
) -> float:
    """Root mean squared calibration error.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: number of discretizations for the probability space [0, 1].
        vectorized: whether to vectorize computation for observed proportions.
                    (while setting to True is faster, it has much higher memory requirements
                    and may fail to run for larger datasets).
        recal_model: an sklearn isotonoic regression model which recalibrates the predictions.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.

    Returns:
        A single scalar which calculates the root mean squared calibration error.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    # Get lists of expected and observed proportions for a range of quantiles
    if vectorized:
        (exp_proportions, obs_proportions) = get_proportion_lists_vectorized(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type
        )
    else:
        (exp_proportions, obs_proportions) = get_proportion_lists(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type
        )

    squared_diff_proportions = np.square(exp_proportions - obs_proportions)
    rmsce = np.sqrt(np.mean(squared_diff_proportions))

    return rmsce


def mean_absolute_calibration_error(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    vectorized: bool = False,
    recal_model: IsotonicRegression = None,
    prop_type: str = "interval",
) -> float:
    """Mean absolute calibration error; identical to ECE.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: number of discretizations for the probability space [0, 1].
        vectorized: whether to vectorize computation for observed proportions.
                    (while setting to True is faster, it has much higher memory requirements
                    and may fail to run for larger datasets).
        recal_model: an sklearn isotonoic regression model which recalibrates the predictions.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.

    Returns:
        A single scalar which calculates the mean absolute calibration error.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    # Get lists of expected and observed proportions for a range of quantiles
    if vectorized:
        (exp_proportions, obs_proportions) = get_proportion_lists_vectorized(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type
        )
    else:
        (exp_proportions, obs_proportions) = get_proportion_lists(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type
        )

    abs_diff_proportions = np.abs(exp_proportions - obs_proportions)
    mace = np.mean(abs_diff_proportions)

    return mace


def adversarial_group_calibration(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    cali_type: str,
    prop_type: str = "interval",
    num_bins: int = 100,
    num_group_bins: int = 10,
    draw_with_replacement: bool = False,
    num_trials: int = 10,
    num_group_draws: int = 10,
    verbose: bool = False,
) -> Namespace:
    """Adversarial group calibration.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        cali_type: type of calibration error to measure; one of ["mean_abs", "root_mean_sq"].
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.
        num_bins: number of discretizations for the probability space [0, 1].
        num_group_bins: number of discretizations for group size proportions between 0 and 1.
        draw_with_replacement: True to draw subgroups that draw from the dataset with replacement.
        num_trials: number of trials to estimate the worst calibration error per group size.
        num_group_draws: number of subgroups to draw per given group size to measure calibration error on.
        verbose: True to print progress statements.

    Returns:
        A Namespace with an array of the group sizes, the mean of the worst
        calibration errors for each group size, and the standard error of the
        worst calibration error for each group size
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    num_pts = y_true.shape[0]

    if cali_type == "mean_abs":
        cali_fn = mean_absolute_calibration_error
    elif cali_type == "root_mean_sq":
        cali_fn = root_mean_squared_calibration_error

    num_pts = y_std.shape[0]
    ratio_arr = np.linspace(0, 1, num_group_bins)
    score_mean_per_ratio = []
    score_stderr_per_ratio = []
    if verbose:
        print(
            (
                "Measuring adversarial group calibration by spanning group"
                " size between {} and {}, in {} intervals"
            ).format(np.min(ratio_arr), np.max(ratio_arr), num_group_bins)
        )
    progress = tqdm(ratio_arr) if verbose else ratio_arr
    for r in progress:
        group_size = max([int(round(num_pts * r)), 2])
        score_per_trial = []  # list of worst miscalibrations encountered
        for _ in range(num_trials):
            group_miscal_scores = []
            for g_idx in range(num_group_draws):
                rand_idx = np.random.choice(
                    num_pts, group_size, replace=draw_with_replacement
                )
                group_y_pred = y_pred[rand_idx]
                group_y_true = y_true[rand_idx]
                group_y_std = y_std[rand_idx]
                group_miscal = cali_fn(
                    group_y_pred,
                    group_y_std,
                    group_y_true,
                    num_bins=num_bins,
                    vectorized=True,
                    prop_type=prop_type,
                )
                group_miscal_scores.append(group_miscal)
            max_miscal_score = np.max(group_miscal_scores)
            score_per_trial.append(max_miscal_score)
        score_mean_across_trials = np.mean(score_per_trial)
        score_stderr_across_trials = np.std(score_per_trial, ddof=1)
        score_mean_per_ratio.append(score_mean_across_trials)
        score_stderr_per_ratio.append(score_stderr_across_trials)

    out = Namespace(
        group_size=ratio_arr,
        score_mean=np.array(score_mean_per_ratio),
        score_stderr=np.array(score_stderr_per_ratio),
    )
    return out


def miscalibration_area(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    vectorized: bool = False,
    recal_model: Any = None,
    prop_type: str = "interval",
) -> float:
    """Miscalibration area.

    This is identical to mean absolute calibration error and ECE, however
    the integration here is taken by tracing the area between curves.
    In the limit of num_bins, miscalibration area and
    mean absolute calibration error will converge to the same value.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: number of discretizations for the probability space [0, 1].
        vectorized: whether to vectorize computation for observed proportions.
                    (while setting to True is faster, it has much higher memory requirements
                    and may fail to run for larger datasets).
        recal_model: an sklearn isotonoic regression model which recalibrates the predictions.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.

    Returns:
        A single scalar which calculates the miscalibration area.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    # Get lists of expected and observed proportions for a range of quantiles
    if vectorized:
        (exp_proportions, obs_proportions) = get_proportion_lists_vectorized(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type=prop_type
        )
    else:
        (exp_proportions, obs_proportions) = get_proportion_lists(
            y_pred, y_std, y_true, num_bins, recal_model, prop_type=prop_type
        )

    # Compute approximation to area between curves
    polygon_points = []
    for point in zip(exp_proportions, obs_proportions):
        polygon_points.append(point)
    for point in zip(reversed(exp_proportions), reversed(exp_proportions)):
        polygon_points.append(point)
    polygon_points.append((exp_proportions[0], obs_proportions[0]))
    polygon = Polygon(polygon_points)
    x, y = polygon.exterior.xy
    ls = LineString(np.c_[x, y])
    lr = LineString(ls.coords[:] + ls.coords[0:1])
    mls = unary_union(lr)
    polygon_area_list = [poly.area for poly in polygonize(mls)]
    miscalibration_area = np.asarray(polygon_area_list).sum()

    return miscalibration_area


def get_proportion_lists_vectorized(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    recal_model: Any = None,
    prop_type: str = "interval",
) -> Tuple[np.ndarray, np.ndarray]:
    """Arrays of expected and observed proportions

    Returns the expected proportions and observed proportion of points falling into
    intervals corresponding to a range of quantiles.
    Computations here are vectorized for faster execution, but this function is
    not suited when there are memory constraints.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: number of discretizations for the probability space [0, 1].
        recal_model: an sklearn isotonoic regression model which recalibrates the predictions.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.

    Returns:
        A tuple of two numpy arrays, expected proportions and observed proportions

    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    # Compute proportions
    exp_proportions = np.linspace(0, 1, num_bins)
    # If we are recalibrating, input proportions are recalibrated proportions
    if recal_model is not None:
        in_exp_proportions = recal_model.predict(exp_proportions)
    else:
        in_exp_proportions = exp_proportions

    residuals = y_pred - y_true
    normalized_residuals = (residuals.flatten() / y_std.flatten()).reshape(-1, 1)
    norm = stats.norm(loc=0, scale=1)
    if prop_type == "interval":
        gaussian_lower_bound = norm.ppf(0.5 - in_exp_proportions / 2.0)
        gaussian_upper_bound = norm.ppf(0.5 + in_exp_proportions / 2.0)

        above_lower = normalized_residuals >= gaussian_lower_bound
        below_upper = normalized_residuals <= gaussian_upper_bound

        within_quantile = above_lower * below_upper
        obs_proportions = np.sum(within_quantile, axis=0).flatten() / len(residuals)
    elif prop_type == "quantile":
        gaussian_quantile_bound = norm.ppf(in_exp_proportions)
        below_quantile = normalized_residuals <= gaussian_quantile_bound
        obs_proportions = np.sum(below_quantile, axis=0).flatten() / len(residuals)

    return exp_proportions, obs_proportions


def get_proportion_lists(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    recal_model: IsotonicRegression = None,
    prop_type: str = "interval",
) -> Tuple[np.ndarray, np.ndarray]:
    """Arrays of expected and observed proportions

    Return arrays of expected and observed proportions of points falling into
    intervals corresponding to a range of quantiles.
    Computations here are not vectorized, in case there are memory constraints.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: number of discretizations for the probability space [0, 1].
        recal_model: an sklearn isotonoic regression model which recalibrates the predictions.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.

    Returns:
        A tuple of two numpy arrays, expected proportions and observed proportions
    """
    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)
    # Check that prop_type is one of 'interval' or 'quantile'
    assert prop_type in ["interval", "quantile"]

    # Compute proportions
    exp_proportions = np.linspace(0, 1, num_bins)
    # If we are recalibrating, input proportions are recalibrated proportions
    if recal_model is not None:
        in_exp_proportions = recal_model.predict(exp_proportions)
    else:
        in_exp_proportions = exp_proportions

    if prop_type == "interval":
        obs_proportions = [
            get_proportion_in_interval(y_pred, y_std, y_true, quantile)
            for quantile in in_exp_proportions
        ]
    elif prop_type == "quantile":
        obs_proportions = [
            get_proportion_under_quantile(y_pred, y_std, y_true, quantile)
            for quantile in in_exp_proportions
        ]

    return exp_proportions, obs_proportions


def get_proportion_in_interval(
    y_pred: np.ndarray, y_std: np.ndarray, y_true: np.ndarray, quantile: float
) -> float:
    """For a specified quantile, return the proportion of points falling into
    an interval corresponding to that quantile.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        quantile: a specified quantile level

    Returns:
        A single scalar which is the proportion of the true labels falling into the
        prediction interval for the specified quantile.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)

    # Computer lower and upper bound for quantile
    norm = stats.norm(loc=0, scale=1)
    lower_bound = norm.ppf(0.5 - quantile / 2)
    upper_bound = norm.ppf(0.5 + quantile / 2)

    # Compute proportion of normalized residuals within lower to upper bound
    residuals = y_pred - y_true

    normalized_residuals = residuals.reshape(-1) / y_std.reshape(-1)

    num_within_quantile = 0
    for resid in normalized_residuals:
        if lower_bound <= resid and resid <= upper_bound:
            num_within_quantile += 1.0
    proportion = num_within_quantile / len(residuals)

    return proportion


def get_proportion_under_quantile(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    quantile: float,
) -> float:
    """Get the proportion of data that are below the predicted quantile.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        quantile: The quantile level to check.

    Returns:
        The proportion of data below the quantile level.
    """

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std, y_true)
    # Check that input std is positive
    assert_is_positive(y_std)

    # Computer lower and upper bound for quantile
    norm = stats.norm(loc=0, scale=1)
    quantile_bound = norm.ppf(quantile)

    # Compute proportion of normalized residuals within lower to upper bound
    residuals = y_pred - y_true

    normalized_residuals = residuals / y_std

    num_below_quantile = 0
    for resid in normalized_residuals:
        if resid <= quantile_bound:
            num_below_quantile += 1.0
    proportion = num_below_quantile / len(residuals)

    return proportion


def get_prediction_interval(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    quantile: np.ndarray,
    recal_model: Optional[IsotonicRegression] = None,
) -> Namespace:
    """Return the centered predictional interval corresponding to a quantile.

    For a specified quantile level q (must be a float, or a singleton),
    return the centered prediction interval corresponding
    to the pair of quantiles at levels (0.5-q/2) and (0.5+q/2),
    i.e. interval that has nominal coverage equal to q.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        quantile: The quantile level to check.
        recal_model: A recalibration model to apply before computing the interval.

    Returns:
        Namespace containing the lower and upper bound corresponding to the
        centered interval.
    """

    if isinstance(quantile, float):
        quantile = np.array([quantile])

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std)
    assert_is_flat_same_shape(quantile)
    assert quantile.size == 1
    # Check that input std is positive
    assert_is_positive(y_std)

    if not np.logical_and((0.0 < quantile.item()), (quantile.item() < 1.0)):
        raise ValueError("Quantile must be greater than 0.0 and less than 1.0")

    # if recal_model is not None, calculate recalibrated quantile
    if recal_model is not None:
        quantile = recal_model.predict(quantile)

    # Computer lower and upper bound for quantile
    norm = stats.norm(loc=y_pred, scale=y_std)
    lower_bound = norm.ppf(0.5 - quantile / 2)
    upper_bound = norm.ppf(0.5 + quantile / 2)

    bounds = Namespace(
        upper=upper_bound,
        lower=lower_bound,
    )

    return bounds


def get_quantile(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    quantile: np.ndarray,
    recal_model: Optional[IsotonicRegression] = None,
) -> float:
    """Return the value corresponding with a quantile.

    For a specified quantile level q (must be a float, or a singleton),
    return the quantile prediction,
    i.e. bound that has nominal coverage below the bound equal to q.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        quantile: The quantile level to check.
        recal_model: A recalibration model to apply before computing the interval.

    Returns:
        The value at which the quantile is achieved.
    """
    if isinstance(quantile, float):
        quantile = np.array([quantile])

    # Check that input arrays are flat
    assert_is_flat_same_shape(y_pred, y_std)
    assert_is_flat_same_shape(quantile)
    assert quantile.size == 1
    # Check that input std is positive
    assert_is_positive(y_std)

    if not np.logical_and((0.0 < quantile.item()), (quantile.item() < 1.0)):
        raise ValueError("Quantile must be greater than 0.0 and less than 1.0")

    # if recal_model is not None, calculate recalibrated quantile
    if recal_model is not None:
        quantile = recal_model.predict(quantile)

    # Computer quantile bound
    norm = stats.norm(loc=y_pred, scale=y_std)
    quantile_prediction = norm.ppf(quantile).flatten()

    return quantile_prediction
